package cs131w.symtab.sto;

//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
//cs131 imports
import  cs131w.error.*;
import  cs131w.assembly.*;
import  cs131w.symtab.types.*;
import	cs131w.parser.MyParser;

public class STO implements Cloneable, Asmbl 
{

	//----------------------------------------------------------------
	// 
	//----------------------------------------------------------------
	protected String  	m_strName;
	protected Type		m_type;
	protected boolean	m_isAddressable;
	protected boolean	m_isModifiable;


	//----------------------------------------------------------------
	// Assembly things
	//----------------------------------------------------------------
	//if this value is in a register somewhere its name is stored here
	//if null it can be found in memory
	protected String		m_register;
	protected String		mem_base;
	protected int			mem_offset;

	//if this object has an address, that can only be determined at runtime (like an array access)
	//this is the storage for the actual address
	//this is to be preferred over the compile-time addr. stored above, if present
	public StorageLocation	mem_runtimeAddr = null;

	//this is the storage location for pointer to struct, if this happens to be a function ptr to a function thats part of a struct
	public StorageLocation	mem_StructThisPtr = null;
	
	
	//unique label to be used for anything in assembly that has to do with this STO
	protected String		m_label;
	
	//if this a member of a struct, if so next var stores that struct's STO for later
	public 	  boolean		m_isStructMember = false;
	public 	  STO			m_memberOfStruct = null;
	
	
	//if this value is not really where it says it is
	//public 	  boolean		m_isVirtual = false;
	
	protected boolean		mem_isLocal = true;
	protected boolean		m_static = false;
	protected boolean		asmbl_created = false;


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public 
	STO ()
	{
		//explicitly set the null type, just in case somebody tries to call functions on the inner type, so it shouldnt be null
		this("", new NullType());
	}

	public 
	STO (String strName)
	{
		//explicitly set the null type, just in case somebody tries to call functions on the inner type, so it shouldnt be null
		this(strName, new NullType());
	}

	public 
	STO (String strName, Type typ)
	{
		setName(strName);
		setType(typ);
		setIsAddressable(false);
		setIsModifiable(false);
		m_register = mem_base = null;
		mem_offset = 0;
		//set the label to something unique, with line number prepended
		m_label = m_strName + "$L" + MyParser.GetLineNum();
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public String
	getName ()
	{
		return m_strName;
	}

	public void
	setName (String str)
	{
		m_strName = str;
	}

	//returns a unique, mangled if needed, label for this STO
	public String
	getLabel ()
	{
		return m_label;
	}

	public void
	setLabel (String l)
	{
		m_label = l;
	}

	//this regenerates the label. Call in case this is global to have label names, that correspond to just the var name
	//if we need to link together with external .o files
	public void
	regenLabel ()
	{
		//if sto is global and not static, fall back to just the name
		if (!mem_isLocal && !m_static)
			m_label = m_strName;
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public Type
	getParentType(){
	//	MyParser.debug("sto.parent() returning " + m_type );
		return m_type;
	}

	public Type
	getType ()
	{
		return	m_type;
	}

	public void
	setType (Type type)
	{
		m_type = type;
	}


	//----------------------------------------------------------------
	// Addressable refers to if the object has an address. Variables
	// and declared constants have an address, whereas results from 
	// expression like (x + y) and literal constants like 77 do not 
	// have an address.
	//----------------------------------------------------------------
	public boolean
	getIsAddressable ()
	{
		return	m_isAddressable;
	}

	public void
	setIsAddressable (boolean addressable)
	{
		m_isAddressable = addressable;
	}

	//----------------------------------------------------------------
	// You shouldn't need to use these two routines directly
	//----------------------------------------------------------------
	public boolean
	getIsModifiable ()
	{
		return	m_isModifiable;
	}

	public void
	setIsModifiable (boolean modifiable)
	{
		m_isModifiable = modifiable;
	}


	//----------------------------------------------------------------
	// A modifiable L-value is an object that is both addressable and
	// modifiable. Objects like constants are not modifiable, so they 
	// are not modifiable L-values.
	//----------------------------------------------------------------
	public boolean
	isModLValue ()
	{
		return	getIsModifiable() && getIsAddressable();
	}

	public void
	setIsModLValue (boolean m)
	{
		setIsModifiable(m);
		setIsAddressable(m);
	}

	public void setRVal (){ setIsModLValue(false); }
	public void setModLVal (){ setIsModLValue(true); }


	public void setStatic (){ m_static = true; }
	public boolean isStatic (){ return m_static; }

	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public String
	toString()
	{
		if (MyParser.DEV){
			String ad = "";
			if (isInReg())
				ad = m_register;
			else
				ad = getAddrStr();
				
			return ("(sto)" + m_strName + ":" + m_type + "@" + ad);
		}else
			return (m_strName);
		
	}
	
	//gets the mangled name for function overloading, just return type info in this case
	public String
	getMangledName()
	{
		return getParentType().getRootType().getMangledName();
	}
	

	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public Object 
	clone() {      //throws CloneNotSupportedException {
		
		Class c = this.getClass();
		STO ret;
		try{
			
			ret = (STO)c.newInstance();
			
		}catch (Exception e){
			
			MyParser.debug(this+".clone: exception occured, returning bare STO! Exception was: ");
			e.printStackTrace();
			ret = new STO();
		
		}
		
		MyParser.debug(this+".clone: setting all fields!");
		ret.setName(new String(m_strName));
		ret.setLabel(new String(m_label));
		ret.setType((Type)m_type.clone());
		
		//STO ret = new STO(new String(m_strName), (Type)m_type.clone());
		
		ret.setIsModifiable(m_isModifiable);
		ret.setIsAddressable(m_isAddressable);

		//now set the asmbly vars, too
		ret.setReg(m_register);
		ret.setMemBase(mem_base);
		ret.setMemOffset(mem_offset);
		ret.setIsLocal(mem_isLocal);
		if (m_static) ret.setStatic();
		//do not set this, since we dont want to overwrite the old asbmly va
		if (asmbl_created) ret.setAsmblyCreated();
		//clone runtime Adress as well
		
		if (hasRuntimeAddr()) ret.setRuntimeAddr((StorageLocation)mem_runtimeAddr.clone());
		
		MyParser.debug(this+".clone: returning " + ret);
		return ret;
	}

	//----------------------------------------------------------------
	//----------------------------------------------------------------
	//general size is the size of the underlying type
	public int
	getSize(){
		return getType().getSize();
	}

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------

	public void setIsStructMember(STO s){
		m_isStructMember = true;
		m_memberOfStruct = s;
	}

	public boolean isStructMember(){
		return m_isStructMember;
	}
	

	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	
	/**
	 * Verifies that constants are known at compile time
	 */
	public STO 
	checkConstExpr(){
		MyParser.debug("checking const " + this + ",of type " + this.getParentType() + ", isFunc:" + this.isFunc());
		if (this.isError() || this.isFunc()) return this;
		
		if(this.isConst()){
			ConstSTO con = (ConstSTO) this;
			if(null != con.getValue()){
				//MyParser.debug("sto is constant, returning!");
				return con;
			}
		}
		
		return new ErrorSTO(this);
		
	}



	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean	isVar () 	{ return false; }
	public boolean	isConst ()	{ return false; }
	public boolean	isExpr ()	{ return false; }
	public boolean	isFunc () 	{ return false; }
	public boolean	isFuncPtr () { return getParentType().isFuncPtr(); }
	public boolean	isTypedef () { return false; }
	public boolean	isError () 	{ return false; }
	public boolean	isParam () 	{ return false; }
	public boolean	isStruct () { return getParentType().isStruct(); }
	public boolean	isEmpty () 	{ return false; }
	public boolean	isNull () 	{ return false; }
	public boolean  isUeberFunc (){ return false;}

	//----------------------------------------------------------------
	// functions for helping with writing assembly
	//----------------------------------------------------------------
	

	//----------------------------------------------------------------
	//	Assembly functions
	//	these emit the necc. assembly code for doing prescribed action
	//----------------------------------------------------------------
	//starts the creation of this STO in assmbly code
	//ie def in case of global Var or prologue in Function
	//or in case of something else, create the value and store it somewhere in temp mem
	public void
	AsmblyCreate(){ 
		AsmblyCreate(null);
	}

	public void
	AsmblyCreate(STO sto){ 
		MyParser.debug(this + ".AsmblyCreate("+sto+") called. OVERRIDE ME!");
		asmbl_created = true;
	}
	
	public void
	setAsmblyCreated(){
		asmbl_created = true;
	}

	public void
	setAsmblyCreated(boolean cr){
		asmbl_created = cr;
	}

	//finish this STO, ie. epilogue in case of function
	public void
	AsmblyFinish(){ 
		MyParser.debug(this + ".AsmblyFinish called. OVERRIDE ME!");
	}

	//loads this STO into a reg, if its in mem
	public void
	AsmblyLoad(){ 
		MyParser.debug(this + ".AsmblyLoad called. OVERRIDE ME!");
	}

	//loads this STO into a reg, if its in mem
	public void
	AsmblyLoad(String into){ 
		MyParser.debug(this + ".AsmblyLoad("+into+") called. OVERRIDE ME!");
	}

	public void
	AsmblyLoadAddr(){
		MyParser.debug(this + ".AsmblyLoadAddr() called. OVERRIDE ME!");
	}

	public void
	AsmblyLoadAddr(String into){
		MyParser.debug(this + ".AsmblyLoadAddr("+into+") called. OVERRIDE ME!");
	}

	//stores this STO back from reg into mem where it was created, and thus frees this reg
	public void
	AsmblyStore(){ 
		MyParser.debug(this + ".AsmblyStore called. OVERRIDE ME!");
	}
	
	public void
	AsmblyStore(String to){ 
		MyParser.debug(this + ".AsmblyStore("+to+") called. OVERRIDE ME!");
	}

	//cleares this STO from reg where it is right now without storing, and thus frees this reg
	public void
	AsmblyClearReg(){ 
		MyParser.debug(this + ".AsmblyClearReg called.");

		//do not store if not addressable or an expression - which lives in temp space (so it is addressable), or if it is an error
		if (isError()) return;
		
		//if not loaded: return
		if (!isInReg()) return;
		
		//and clear
		Regs.FreeReg(m_register);
		m_register = null;
	
	}


	public void
	AsmblyCOut(){ 
		MyParser.debug(this + ".AsmblyCOut called. OVERRIDE ME!");
	}

	public void
	AsmblyCIn(){ 
		MyParser.debug(this + ".AsmblyCIn called. OVERRIDE ME!");
	}

	
	public String
	getRegister(){
		return m_register;
	}

	public String
	getReg(){
		return m_register;
	}

	public void
	setRegister(String r){
		m_register = r;
	}

	public void
	setReg(String r){
		m_register = r;
	}

	public boolean
	isInReg(){
		return (m_register != null);
	}
	
	public String
	getMemBase(){
		return mem_base;
	}

	public void
	setMemBase(String base){
		mem_base = base;
	}

	public int
	getMemOffset(){
		return mem_offset;
	}

	public void
	setMemOffset(int off){
		mem_offset = off;
	}

	public String
	getAddrStr(){
		String o = "";
		if (mem_offset >= 0) o = "+";
		o += mem_offset;
		
		return (mem_base + o);
	}

	public boolean
	isLocal(){
		return mem_isLocal;
	}

	public boolean
	hasRuntimeAddr(){
		return (mem_runtimeAddr != null);
	}

	public void
	setRuntimeAddr(StorageLocation loc){
		mem_runtimeAddr = loc;
	}

	public void
	setStructThisPtr(StorageLocation loc){
		mem_StructThisPtr = loc;
	}

	public boolean
	hasStructThisPtr(){
		return (mem_StructThisPtr != null);
	}

	public void
	loadRuntimeAddrInto(String into){
		MyParser.debug(this + ".loadRuntimeAddrInto("+into+") called");
		
		if (mem_runtimeAddr == null) return;
		
		//if already loaded: return
		if (mem_runtimeAddr.isInReg() && mem_runtimeAddr.getReg().equals(into)){
			MyParser.debug(this + ".loadRuntimeAddrInto("+into+") already loaded in "+into+", returning");
			return;
		}
		
		//if loaded but somewhere else, move it
		if (mem_runtimeAddr.isInReg() && mem_runtimeAddr.getReg() != into){
		 	MyParser.debug(this + ".loadRuntimeAddrInto("+into+") moving from " + mem_runtimeAddr.getReg());
			//need to move it there
			MyParser.asmblyWriteVarSmt(MOVE, mem_runtimeAddr.getReg(), into);
			
			//and free old one if it happens to be local
			Regs.FreeReg(mem_runtimeAddr.getReg());
			mem_runtimeAddr.setReg(into);
			return;
		}
		
		MyParser.debug(this + ".loadRuntimeAddrInto writing load from: " + mem_runtimeAddr.getAddrStr());
		//is always local, so load it
		MyParser.asmblyWriteVarSmtComm ("Load '" + this + "' runtime addr into " + into, LOAD, mem_runtimeAddr.getAddrStr(), into);
		mem_runtimeAddr.setReg(into);
	}
	

	public void
	setIsLocal(boolean l){
		mem_isLocal = l;
		regenLabel();
	}
	
	//mark this as global
	public void
	markGlobal(){
		mem_offset = 0;
		mem_isLocal = false;
		regenLabel();
		mem_base = getLabel();
	}

	//mark this as a member of a struct
	public void
	markStructMember(int off){
		mem_base = "struct";
		mem_offset = off;
		mem_isLocal = false;
		regenLabel();
	}

	//mark this as local to a func
	public void
	markLocal(int offs){
		mem_base = FP;
		mem_offset = offs;
		mem_isLocal = true;
		regenLabel();
	}

	//mark this as extern, meaning just set its label, without actually creating it
	public void
	markExtern(){
		asmbl_created = true;
		mem_offset = 0;
		mem_isLocal = false;
		regenLabel();
		mem_base = getLabel();
	}



}
