package cs131w.symtab.types;
import 	java.util.Vector;
import 	cs131w.symtab.sto.*;
import 	cs131w.error.*;
import 	cs131w.assembly.*;
import 	cs131w.parser.MyParser;


public class FuncProtoType extends Type implements Asmbl
{

	//----------------------------------------------------------------
	//	Instance variables.
	//----------------------------------------------------------------
		//the return base type
		private Type 		m_returnType;
		//a vector of ParamSTOs denoting the params
		protected Vector	m_VParams;
	
	
	//----------------------------------------------------------------
	// this is the the prototype that completely defines a function
	//----------------------------------------------------------------
	public 
	FuncProtoType ()
	{
		this(new VoidType());
	}


	public 
	FuncProtoType (Type ret)
	{
		super("function", 0);
		m_VParams = new Vector();
		m_returnType = ret;
	}


	//ptrs actually have state, so override clone
	public Object 
	clone(){
		FuncProtoType ret = new FuncProtoType((Type)m_returnType.clone());
		ret.setParams((Vector)m_VParams.clone());
		return ret;
	}


	//----------------------------------------------------------------
	//
	//----------------------------------------------------------------
	public void
	addParam(ParamSTO p){
		m_VParams.addElement(p);
	}

	public ParamSTO
	getParamByName(String strName){
		ParamSTO		sto = null;
		
		for (int i = 0; i < m_VParams.size(); i++)
		{
			sto = (ParamSTO) m_VParams.elementAt (i);
			if (sto.getName().equals (strName))
				return (sto);
		}

		return (null);
	}

	
	public ParamSTO
	getParam(int i){
		return (ParamSTO) m_VParams.elementAt (i);
	}

	public Vector
	getParams(){
		return m_VParams;
	}

	public void
	setParams(Vector v){
		m_VParams = v;
	}

	//----------------------------------------------------------------
	// This is the return type of the function. 
	//----------------------------------------------------------------
	public void
	setReturnType (Type typ)
	{
		m_returnType = typ;
	}

	public Type
	getReturnType ()
	{
		return m_returnType;
	}


	//----------------------------------------------------------------
	//----------------------------------------------------------------

	public int
	getNumParams ()
	{
		return m_VParams.size();
	}



	//----------------------------------------------------------------
	// gets a string describing this function Proto
	//----------------------------------------------------------------
	public String
	toString()
	{
		final StringBuilder sb = new StringBuilder();
		
		if (MyParser.DEV)
			sb.append("function : ");
		else
			sb.append("funcptr : ");
		
		sb.append(m_returnType + " (");	
	
		ParamSTO sto;
		for (int i = 0; i < m_VParams.size (); i++)
		{
			sto = (ParamSTO) m_VParams.elementAt (i);
			sb.append( ((i>0) ? ", " : "") + sto);
		}
		
		sb.append(")");
		return sb.toString();
		
	}	

	//----------------------------------------------------------------
	// gets a unique mangled name for this proto
	//----------------------------------------------------------------
	public String
	getMangledName()
	{
		final StringBuilder sb = new StringBuilder();
		ParamSTO sto;
		//append each param in order
		for (int i = 0; i < m_VParams.size (); i++)
		{
			sto = (ParamSTO) m_VParams.elementAt (i);
			sb.append( "_$" + sto.getMangledName());
		}
		
		MyParser.debug(this + ".getMangledName() returning: '" + sb + "'");
		return sb.toString();
	}

	//----------------------------------------------------------------
	// equivalent checks the actual function protos against each other
	//----------------------------------------------------------------
	public boolean isEquivalent(Type t){
		
		//MyParser.debug("func-proto checking equiv: "+this+" == " + t);
		
		if (t == null || !t.isFunction() ) return false;
		
		FuncProtoType f = (FuncProtoType) t.getRootType();
		
		//now check the params
		if ( 
				(f.getNumParams() != this.getNumParams()) 
			|| 
				!( this.getReturnType().getRootType().isEquivalent(f.getReturnType().getRootType()) ) 
			){
				//MyParser.debug("#params differ");
				return false;
				
			}
		
		Vector others = f.getParams();

		ParamSTO		sto = null;
		ParamSTO		sto2 = null;
		
		for (int i = 0; i < m_VParams.size (); i++)
		{
			sto = (ParamSTO) m_VParams.elementAt (i);
			sto2 = (ParamSTO) others.elementAt (i);

			//check ref status and type only, not name!
			if ( (sto.isRef() != sto2.isRef() ) || !(sto.getParentType().isEquivalent(sto2.getParentType())) ){
				//MyParser.debug("param "+i+" differs, this: "+sto+ ",o: "+sto2);
				return false;
			}
		}

		return true;
	}

	public boolean isAssignable(Type t){
		//first check if t is a funcptr
		if (t.getRootType().isFuncPtr()){
			return this.isEquivalent(t.getRootType().getPointedToType());
		}
		
		//override assignable to signal if the return type is assignable to type t
		return (m_returnType.getRootType().isAssignable(t.getRootType()));
	}	
	

	public static
	FuncProtoType create(Vector params, Type ret){
		
		FuncProtoType func = new FuncProtoType(ret);
		
		// insert parameters
		for (int i = 0; i < params.size (); i++)
		{
			ParamSTO var = (ParamSTO) (params.elementAt(i));
			//check if already there
			if (null != func.getParamByName(var.getName())){
				
				//return new ErrorType (Formatter.toString(ErrorMsg.redeclared_id, var.getName() ));
				MyParser.STO_doErrorReport (var, Formatter.toString(ErrorMsg.redeclared_id, var.getName() ));
				continue;
			}
			
			//add to function as param
			func.addParam(var);
			MyParser.debug("adding var '"+var+"' to params of func '" + func + "'");
		}
		
		return func;	
		
	}


	//----------------------------------------------------------------
	// parses a function call to this with params p,
	// which is a vector of types STO 
	//----------------------------------------------------------------
	public STO
	DoFuncCall (Vector p)
	{
		MyParser.debug (this + ".funcall calling with " + p);
		boolean bError = false;
		STO ret = new EmptySTO();
		
		//do num params check
		if (p.size() != getNumParams()){
			return MyParser.STO_doErrorReport (Formatter.toString(ErrorMsg.error5n_Call, p.size(), getNumParams()));
		}
		
		//allow mult. errors per line
		MyParser.m_allowDupErrorLine = true;
		
		//now do the params check
		for (int i = 0; i < p.size (); i++)
		{
			STO arg = (STO) p.elementAt (i);
			STO res = null;
			ParamSTO param = getParam(i);
			
			MyParser.debug (this + ".funcall checking " + arg + ".assignable to " + param);
			//check if arg is assignable to param
			if (! arg.getType().isAssignable(param.getType())){
				//is the arg a ref param or a value param?
				if(param.getIsRef()){
					ret = MyParser.STO_doErrorReport (arg, Formatter.toString(ErrorMsg.error5r_Call, arg.getType(), param.getName(), param.getType()));
				}else{
					ret = MyParser.STO_doErrorReport (arg, Formatter.toString(ErrorMsg.error5a_Call, arg.getType(), param.getName(), param.getType()));
				}
				
				bError = true;
				continue;
			} 
			
			//check if ref
			if (param.getIsRef()){
				//and equivalency
				if (! arg.getType().isEquivalent(param.getType())){
					ret = MyParser.STO_doErrorReport (arg, Formatter.toString(ErrorMsg.error5r_Call, arg.getType(), param.getName(), param.getType()));
					bError = true;
					continue;
				}
				//check mod L-val
				if (!arg.isModLValue()){
					if(!arg.getType().isArray()){
						ret = MyParser.STO_doErrorReport (arg, Formatter.toString(ErrorMsg.error5c_Call, param.getName(), arg.getType()));
						bError = true;
						continue;
					}
				}
			}
		}
		
		MyParser.m_allowDupErrorLine = false;
		//if no error make new exprSTO with appr. return type
		if (!bError)
			ret = new ExprSTO("func-call", getReturnType());
			
		//if theres an error, just return last error
		return ret;
		
	}

	//----------------------------------------------------------------
	//----------------------------------------------------------------
	public boolean	isFunction ()	{ return true; }
	
	//----------------------------------------------------------------
	//----------------------------------------------------------------

	//this does the function call in asmbly, set all params and call
	//p is Vector of actual args, ret return STO, struct_this current struct if any,
	//p_call is label or addr to call
	public void
	AsmblyCall(Vector p, STO ret, STO struct_this, String p_call){
		
		String strptr = null;
		
		if (struct_this != null){
			struct_this.AsmblyLoadAddr(O0);
			strptr = O0;
		}
		
		this.AsmblyCall(p, ret, strptr, p_call);
	}	
	
	//this does the function call in asmbly, set all params and call
	//p is Vector of actual args, ret return STO, struct_this is Reg that current structPtr is in
	//p_call is label or addr to call
	public void
	AsmblyCall(Vector p, STO ret, String struct_this, String p_call){
		
		int start_reg = 0;
		int start_stack = 92;
		
		//if inside of struct, Addr of it 'this' should be in o0
		if (struct_this != null){
			//struct_this.AsmblyLoadAddr(O0);
			start_reg = 1;
		}
		
		STO arg;
		ParamSTO parm;
		
		int n_args = p.size() + start_reg;
		String add_sp = "";
		boolean more_six = (n_args > 6);
		
		//check to see if we have more than 6 params, we need to save some additional space
		if (more_six){
			add_sp = "-(" + (n_args-6) + " *4) & -8";
			MyParser.asmblyWriteVarSmtComm ("save add. space on stack for " + (n_args-6) + " params", ADD, SP, add_sp, SP);
		}
		
		for (int i = 0; i < p.size(); i++)
		{
			arg = (STO) p.elementAt (i);
			parm = getParam(i);
			 
			//see if we are over 6
			if (start_reg > 5){
				//in this case we need to load into %fp + start_stack
				//load first
				if (parm.getIsRef() || parm.getType().isStruct() || parm.getType().isArray()){
					arg.AsmblyLoadAddr(L7);
					MyParser.asmblyWriteVarSmt (STORE, L7, SP + "+" + start_stack);
					
				}else{
					arg.AsmblyLoad();
					arg.AsmblyStore(SP + "+" + start_stack);
				}	
				//and store back to stack addr
				start_stack += 4;
				
			}else{
			
				//check if the corresp. param is pass-by-ref, array or struct
				if (parm.getIsRef() || parm.getType().isStruct() || parm.getType().isArray()){
					//load the addr into out reg
					arg.AsmblyLoadAddr("%o" + start_reg);
				}else{
					if (arg.getParentType().isFloat()){
						arg.AsmblyLoad("%f" + start_reg);
					}else{
						arg.AsmblyLoad("%o" + start_reg);
					}
			
				}
			}
			
			//and clear param from any regs, if it's loaded, to make sure it gets loaded fresh from mem next time
			//NOTE: this does not store back, or emit any asmbly code, if param was loaded, it will stay in reg
			arg.AsmblyClearReg();
			start_reg++;
		}
		
		
		//now call
		MyParser.asmblyWriteVarSmt ("\n");
		
		MyParser.asmblyWriteVarSmtComm ("Do the actual call", CALL, p_call);
		MyParser.asmblyWriteVarSmtComm ("NoOp in delay slot", NOP);
		
		MyParser.asmblyWriteVarSmt ("\n");
		
		//now do post-return
		//deallocate stack space, if more than 6
		if (more_six){
			MyParser.asmblyWriteVarSmtComm("and deallocate stack space", SUB, SP, add_sp, SP);
		}
	
		
		//first move return param into appr. reg
		if (!getReturnType().isVoid()){
			//check if ret type is float 
			if (getReturnType().isFloat()){
				//and signal that the return value is in f0
				ret.setReg(F0);
			}else{
				//move from o0 to l0
				MyParser.asmblyWriteVarSmt (MOVE, O0, L0);
				//and signal that the return value is in l0
				ret.setReg(L0);
			}
		}
		
	}
	
	
}